/*
 * Copyright (C) 2006-2007	Andre Beckedorf
 * 					 		<evilJazz _AT_ katastrophos _DOT_ net>
 *
 * The few remaining parts in this almost completely rewritten file are
 * Copyright (C) 2005 Atmark <atmarkat _AT_ msn _DOT_ com>
 *                    AGAWA Koji <i _AT_ atty _DOT_ jp>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <cstdlib>
#include <ctime>
#include <time.h>

#include <qstring.h>
#include <qapplication.h>
#include <qdatetime.h>

#ifdef QT4
#include <classcompathack.h>
#else
#ifdef QTOPIA
#include <qpe/config.h>
#include "compat/sharp-qtopia/qt/qlistview.h"
#include <qtextstream.h>
#include <qheader.h>
#else
#include <config.h>
#include <qtextstream.h>
#include <qheader.h>
#include <qlistview.h>
#endif
#endif

#include <qpainter.h>
#include <qimage.h>
#include "mediadatabase.h"

#include "media.h"
#include "playlistview.h"
#include "debug.h"
#include "skinmanager.h"
#include "sqlite3.h"

#include "compathack.h"

/* PlayListViewItem */

PlayListViewItem::PlayListViewItem(PlayListAspectItem *item, QListView *parent)
	: QListViewItem(parent),
	  item_(item)
{
}

PlayListViewItem::PlayListViewItem(PlayListAspectItem *item, QListView *parent, QListViewItem *after)
	: QListViewItem(parent, after),
	  item_(item)
{
}

QString PlayListViewItem::key(int column, bool) const
{
	// omit sorting. We'll do it in our DB backend...
	return NULL;
}

void PlayListViewItem::paintFocus(QPainter *, const QColorGroup &, const QRect &)
{
	// don't paint focus rect...
}

void PlayListViewItem::paintCell(QPainter * p, const QColorGroup & cg, int column, int width, int align )
{
    if (!p)
    	return;

	int ownHeight = height();
	PlayListView *lv = static_cast<PlayListView *>(listView());

	int vy = lv->itemRect(this).y();   // get y inside the viewport
	int x, y;

	lv->viewportToContents(0, vy, x, y);  // get y inside the contents

	QColorGroup grp;

	if (ownHeight != 0 && (y / ownHeight) % 2 == 0)
		grp = lv->evenRowColorGroup_;
	else
		grp = lv->oddRowColorGroup_;

	p->fillRect( 0, 0, width, ownHeight, grp.brush(QColorGroup::Base));

	int r = 1;
	int marg = 1;

	if (isSelected())
	{
		p->fillRect(r - marg, 0, width - r + marg, ownHeight,
					cg.brush(QColorGroup::Highlight));

		p->setPen(cg.highlightedText());
	}
	else
	{
		p->setPen(cg.text());
	}

	if (column == 0)
	{
		if (this == lv->activeItem_)
		{
			const QPixmap *indicator = lv->activeItemPaused_ ? lv->pauseIndicatorImage_ : lv->playIndicatorImage_;
			if (indicator)
			{
				int t = (ownHeight - indicator->height()) / 2;
				p->drawPixmap(r, t, *indicator);
				r += indicator->width();
			}
		}

		if (this->media()->errorDetected() && lv->errorIndicatorImage_)
		{
			const QPixmap *indicator = lv->errorIndicatorImage_;
			int t = (ownHeight - indicator->height()) / 2;
			p->drawPixmap(r, t, *indicator);
		}
	}
	else
	{
		QString t;
		Media *currentMedia = media();

		switch (column)
		{
		case 0:
			t = "    ";
			break;
		case 1:
			t = currentMedia->trackNo() + "  ";
			break;
		case 2:
			t = currentMedia->title();
			break;
		case 3:
			t = currentMedia->artist();
			break;
		case 4:
			t = currentMedia->album();
			break;
		case 5:
			t = currentMedia->playTime();
			break;
		}

		if (!t.isEmpty())
			p->drawText(r, 0, width - marg - r, ownHeight, align, t);
	}
}

int PlayListViewItem::width( const QFontMetrics& fm, const QListView* lv, int c ) const
{
	return lv->columnWidth(c);
}

void PlayListViewItem::setSelected(bool selected)
{
	QListViewItem::setSelected(selected);
	static_cast<PlayListView *>(this->listView())->itemSelectedStateChanged(this, selected);
}

/* PlayListView */

PlayListView::PlayListView(QWidget *parent)
	: KineticListView(parent),
		mediaDatabase_(NULL),
		aspectSource_(NULL),
		userIsResizingSection_(false),
		autoResizingSections_(false),
		updateCount_(0),
		updatingView_(false),
		playOrder_(ORDER_NORMAL),
		playOrderTable(NULL),
		count_(0), currentIndex(0),
		evenRowColorGroup_(colorGroup()),
		oddRowColorGroup_(colorGroup()),
		playIndicatorImage_(NULL),
		pauseIndicatorImage_(NULL),
		errorIndicatorImage_(NULL),
		activeItem_(NULL),
		activeID_(0),
		activeMediaID_(0),
		activeItemPaused_(false),
		inputMode_(Select),
		directMode_(false),
		dbPlaylistItemOrderDirty_(false),
		stmtInsertMediaIDIntoPlaylist(NULL)
{
	// set to manual width mode.
	// this boosts performance quite dramatically as the
	// column width doesn't need to be measured
	addColumn(tr("   "), 20);
	setColumnAlignment(addColumn(tr("No."), 100), Qt::AlignRight);
	addColumn(tr("Title"), 100);
	addColumn(tr("Artist"), 100);
	addColumn(tr("Album"), 100);
	setColumnAlignment(addColumn(tr("Length"), 40), Qt::AlignRight);

	setRootIsDecorated(false);
	setAllColumnsShowFocus(true);

#ifdef QTOPIA
	setWFlags(getWFlags() | Qt::WNorthWestGravity | Qt::WRepaintNoErase | Qt::WResizeNoErase);
#else
#ifndef QT4
	setWFlags(getWFlags() | Qt::WStaticContents | Qt::WNoAutoErase);
#endif
#endif

	setInputMode(PlayListView::Select);

	QListView::setSorting(-1);

	header()->setClickEnabled(true, 1);
	header()->setClickEnabled(true, 2);
	header()->setClickEnabled(true, 3);
	header()->setClickEnabled(true, 4);
	header()->setClickEnabled(true, 5);

	header()->setResizeEnabled(false, 0);

	connect(header(), SIGNAL(indexChange(int, int, int)),
			this, SLOT(headerIndexChange(int, int, int)));

	connect(header(), SIGNAL(clicked(int)),
			this, SLOT(headerClicked(int)));

	connect(header(), SIGNAL(sizeChange(int, int, int)),
			this, SLOT(headerSizeChange(int, int, int)));

	setFrameStyle(QFrame::NoFrame);

	connect(this, SIGNAL(doubleClicked(QListViewItem *)), this, SIGNAL(playbackIntent()));

	setAutoScaleColumnsEnabled(true);
}

PlayListView::~PlayListView()
{
	QListView::clear();
}

void PlayListView::attachTo(PlayListAspectListViewRenderer *source)
{
	detach();

	if (source)
	{
		aspectSource_ = source;
		mediaDatabase_ = aspectSource_->playListAspect()->playList()->mediaDatabase();

		connect(aspectSource_, SIGNAL(updateView()), this, SLOT(updateView()));

		connect(aspectSource_->playListAspect()->playList()->mediaDatabase(), SIGNAL(preparingStatements(sqlite3*)),
				this, SLOT(mediaDatabasePreparingStatements(sqlite3*)));

		connect(aspectSource_->playListAspect()->playList()->mediaDatabase(), SIGNAL(finalizingStatements(sqlite3*)),
				this, SLOT(mediaDatabaseFinalizingStatements(sqlite3*)));

		header()->setSortIndicator(aspectSource_->sortedColumn(), !aspectSource_->isSortAscending());

		updateView();

		setCurrentItem(findItemByIndex(aspectSource_->selectedItemIndex()));
		setContentsPos(aspectSource_->viewportPosition().x(), aspectSource_->viewportPosition().y());
	}
}

void PlayListView::detach()
{
	ensurePlayListValidInDB();

	if (aspectSource_)
	{
		aspectSource_->setSelectedItemIndex(currentItem() ? itemIndex(currentItem()) : -1);
		aspectSource_->setViewportPosition(QPoint(contentsX(), contentsY()));

		disconnect(aspectSource_->playListAspect()->playList()->mediaDatabase(), SIGNAL(finalizingStatements(sqlite3*)),
				this, SLOT(mediaDatabaseFinalizingStatements(sqlite3*)));

		disconnect(aspectSource_->playListAspect()->playList()->mediaDatabase(), SIGNAL(preparingStatements(sqlite3*)),
				this, SLOT(mediaDatabasePreparingStatements(sqlite3*)));

		disconnect(aspectSource_, SIGNAL(updateView()), this, SLOT(updateView()));

		activeItem_ = NULL;
		selectedItems_.clear();
		QListView::clear();
		dbPlaylistItemOrderDirty_ = false;

		mediaDatabase_ = NULL;
		aspectSource_ = NULL;
	}
}

/*
void PlayListView::clear(bool clearDB)
{
	beginUpdate();
	activeItem_ = NULL;
	activeID_ = 0;
	activeMediaID_ = 0;
	selectedItems_.clear();
	QListView::clear();

	if (aspectSource_ && mediaDatabase_)
	{
		mediaDatabase_->clearCache();
		if (clearDB)
			aspectSource_->playListAspect()->playList()->clearTable();
	}

	dbPlaylistItemOrderDirty_ = true;

	endUpdate();
}
*/

void PlayListView::beginUpdate()
{
	DENTERMETHOD;

	if (updateCount_ == 0)
	{
		viewport()->setUpdatesEnabled(false);
		setUpdatesEnabled(false);
		blockSignals(true);

		/*
		if (mediaDatabase_)
			mediaDatabase_->beginUpdate();
		*/
	}
	++updateCount_;

	DEXITMETHOD;
}

void PlayListView::endUpdate()
{
	DENTERMETHOD;

	--updateCount_;
	if (updateCount_ == 0)
	{
		DTIMERINIT(time);

		/*
		if (mediaDatabase_)
		{
			mediaDatabase_->endUpdate();
			DTIMERPRINT(time, "mediaDatabase_->endUpdate()");
		}
		*/

		DTIMERSTART(time);
		blockSignals(false);
		setUpdatesEnabled(true);
		viewport()->setUpdatesEnabled(true);
		triggerUpdate();
		DTIMERPRINT(time, "list view refresh");
	}

	DEXITMETHOD;
}

void PlayListView::beginChangingUpdate()
{
	DENTERMETHOD;

	beginUpdate();

	// switch to unsorted, unfiltered mode...

	// TODO: beginChangingUpdate() in PlayList
	/*
	savedState = state;
	state.isSorted = false;
	state.isFiltered = false;
	state.filterText = "";

	// direct mode means external filtering will be omitted
	// the listview's content and order is the same as in DB table "playlist" after calling updateView();
	directMode_ = true;

	updateView();
	*/

	DEXITMETHOD;
}

void PlayListView::endChangingUpdate()
{
	DENTERMETHOD;

	// TODO: endChangingUpdate() in PlayList
	/*
	state = savedState;
	directMode_ = false;
	*/

	endUpdate();

	DEXITMETHOD;
}

void PlayListView::ensurePlayListValidInDB()
{
	if (!aspectSource_ || !mediaDatabase_) return;

	DENTERMETHOD;
	if (!aspectSource_->isGraphFiltered() && dbPlaylistItemOrderDirty_)
	{
		DPRINTF("playlist item order is dirty, we have to save the order to the DB...");
		transferPlayListToDB();
	}
	DEXITMETHOD;
}

void PlayListView::transferPlayListToDB()
{
	if (!aspectSource_ || !mediaDatabase_) return;

	DENTERMETHOD;

	DPRINTF("aspectSource_->isGraphFiltered() : %s", aspectSource_->isGraphFiltered() ? "true" : "false");
	DPRINTF("dbPlaylistItemOrderDirty_ : %s", dbPlaylistItemOrderDirty_ ? "true" : "false");
	if (aspectSource_->isGraphFiltered())
	{
		DEXITMETHOD;
		return;
	}

	beginUpdate();

	aspectSource_->playListAspect()->playList()->clearTable();

	PlayListViewItem *item;
	int i = 0;

	sqlite3_stmt *vm = stmtInsertMediaIDIntoPlaylist;

	QListViewItemIterator it(this);
	while (it.current())
	{
		item = static_cast<PlayListViewItem *>(it.current());

		if (!item->id())
		{
			// set new playlist ID if the playlist ID was previously NULL
			sqlite3_bind_int(vm, 1, i);
			sqlite3_bind_null(vm, 2);
			sqlite3_bind_int(vm, 3, item->mediaID());
			sqlite3_step(vm);
			item->setID(sqlite3_last_insert_rowid(mediaDatabase_->db_));
			DHPRINTF("new item id: %d", item->id());
		}
		else
		{
			DHPRINTF("item id: %d", item->id());
			sqlite3_bind_int(vm, 1, i);
			sqlite3_bind_int(vm, 2, item->id());
			sqlite3_bind_int(vm, 3, item->mediaID());
			sqlite3_step(vm);
			DHOP(mediaDatabase_->dbDebug("sqlite3_step(vm);"));
		}

		DOP(mediaDatabase_->dbDebug("stmtInsertMediaIDIntoPlaylist " + item->mediaID()));
		sqlite3_reset(vm);

		++it;
		++i;
	}

	endUpdate();

	dbPlaylistItemOrderDirty_ = false;

	DEXITMETHOD;
}

void PlayListView::updateView()
{
	DENTERMETHOD;

	beginUpdate();

	bool dbPlaylistItemOrderWasDirty = dbPlaylistItemOrderDirty_;
	ensurePlayListValidInDB();

	unsigned long curActiveID = activeID_;
	unsigned long curActiveMediaID = activeMediaID_;
	// save ID of currently selected item so we can select it again later on
	unsigned long curSelectedID = 0;
	if (currentItem())
		curSelectedID = static_cast<PlayListViewItem *>(currentItem())->id();

	activeItem_ = NULL;
	activeID_ = 0;
	activeMediaID_ = 0;

	DTIMERINIT(time);

	QListView::clear();

	DTIMERPRINT(time, "graphExecuted ht update");

	header()->setSortIndicator(aspectSource_->sortedColumn(), !aspectSource_->isSortAscending());

	PlayListViewItem *lastNewItem = NULL;
	PlayListAspectItemList items = aspectSource_->itemList();
	for (int i = 0; i < items.count(); ++i)
		lastNewItem = new PlayListViewItem(items.at(i), this, lastNewItem);

	DTIMERPRINT(time, "graphExecuted t update");

	// try to select the previously selected item...
	DPRINTF("setActiveItemByID(%d);", curActiveID);
	if (setActiveItemByID(curActiveID))
	{
		// something was found by the playlist ID!
		// now check, if the media ID fits too.
		// if not, try to set by media ID...
		if (activeItem()->mediaID() != curActiveMediaID)
		{
			setActiveItemByMediaID(curActiveMediaID);
		}
	}
	else
	{
		// nothing was found by playlist ID, try with
		// media ID instead...
		setActiveItemByMediaID(curActiveMediaID);
	}

	// even if nothing was found, force activeID...
	activeID_ = curActiveID;
	activeMediaID_ = curActiveMediaID;

	// try to find previous selection, reselect it and make it visible
	if (curSelectedID)
	{
		DPRINTF("PlayListItem *foundItem = findItemByID(curSelectedID);");
		PlayListViewItem *foundItem = findItemByID(curSelectedID);
		if (foundItem)
		{
			DPRINTF("setCurrentItem(foundItem);");
			setCurrentItem(foundItem);
			DPRINTF("ensureItemVisible(foundItem);");
			ensureItemVisible(foundItem);
		}
	}

	// make sure to select at least something...
	if (!currentItem() && firstChild())
	{
		DPRINTF("setCurrentItem(firstChild());");
		setCurrentItem(firstChild());
		DPRINTF("ensureItemVisible(firstChild());");
		ensureItemVisible(firstChild());
	}

	endUpdate();

	if (autoScaleColumnsEnabled_)
		setAutoSectionSizes();

	DEXITMETHOD;
}

void PlayListView::setInputMode(PlayListView::InputMode inputmode)
{
	switch (inputmode)
	{
	case PlayListView::Select:
		selectAll(false);
		setSelectionMode(QListView::Single);
		if (currentItem())
			currentItem()->setSelected(true);
		break;
	case PlayListView::MultiSelect:
	case PlayListView::Move:
		setSelectionMode(QListView::Extended);
		break;
	}

	inputMode_ = inputmode;
}

void PlayListView::takeItem(QListViewItem *item)
{
	if (item == activeItem_)
		setActiveItem(0);

	dbPlaylistItemOrderDirty_ = true;

	QListView::takeItem(item);
}

MediaIDList PlayListView::getSelectedMediaIDs()
{
	MediaIDList list;
	for (int i = 0; i < selectedItems_.count(); ++i)
		list.append(selectedItems_.at(i)->mediaID());
	return list;
}

MediaList PlayListView::getSelectedMedia()
{
	MediaList list;
	for (int i = 0; i < selectedItems_.count(); ++i)
		list.append(selectedItems_.at(i)->media());
	return list;
}

PlayListIDList PlayListView::getSelectedItemIDs()
{
	PlayListIDList list;
	for (int i = 0; i < selectedItems_.count(); ++i)
		list.append(selectedItems_.at(i)->id());
	return list;
}

PlayListViewItemList & PlayListView::selectedItems()
{
	return selectedItems_;
}

void PlayListView::removeSelectedItems()
{
	if (selectedItems_.count() > 0)
	{
		// important to get the selected IDs here, since beginChangingUpdate
		// will clear the list and thus the selection...
		PlayListIDList itemIDs = getSelectedItemIDs();

		beginChangingUpdate();

		aspectSource_->playListAspect()->playList()->deletePlayListIDs(itemIDs);

		endChangingUpdate();
		//reloadView();
	}
}

void PlayListView::headerIndexChange(int section, int fromIndex, int toIndex)
{
	// don't allow moving the first column...
	if (header()->mapToIndex(0) != 0)
	{
		header()->moveSection(0, 0); // make sure, blank column is always first.
		header()->update();
	}

	updateResizabilityOfSections();
}

void PlayListView::updateDynamicSectionSizes()
{
	float totalScalingSectionSize = 0;

	totalScalingSectionSize += header()->sectionSize(2);
	totalScalingSectionSize += header()->sectionSize(3);
	totalScalingSectionSize += header()->sectionSize(4);

	headerSectionSizes_.clear();

	headerSectionSizes_.append((float)header()->sectionSize(2) / totalScalingSectionSize * 100);
	headerSectionSizes_.append((float)header()->sectionSize(3) / totalScalingSectionSize * 100);
	headerSectionSizes_.append((float)header()->sectionSize(4) / totalScalingSectionSize * 100);
}

void PlayListView::headerSizeChange(int section, int oldSize, int newSize)
{
	if (!autoScaleColumnsEnabled_ || userIsResizingSection_ || autoResizingSections_)
		return;

	qDebug("PlayListView::headerSizeChange(int section, int oldSize, int newSize): %d, %d, %d", section, oldSize, newSize);

	userIsResizingSection_ = true;

	int index = header()->mapToIndex(section);
	int sizeDiff = newSize - oldSize;

	int nextIndex = index + 1;
	int neighborSection = 0;
	while (nextIndex < header()->count())
	{
		neighborSection = header()->mapToSection(nextIndex);
		if (neighborSection == 2 || neighborSection == 3 || neighborSection == 4 || nextIndex == header()->count() - 1)
			break;

		neighborSection = 0;
		++nextIndex;
	}

	if (neighborSection != 0)
		setColumnWidth(neighborSection, columnWidth(neighborSection) - sizeDiff);

	updateDynamicSectionSizes();

	userIsResizingSection_ = false;
}

void PlayListView::setAutoScaleColumnsEnabled(bool value)
{
	autoScaleColumnsEnabled_ = value;

	if (autoScaleColumnsEnabled_)
	{
		setHScrollBarMode(QListView::AlwaysOff);
		setAutoSectionSizes();
	}
	else
	{
		updateResizabilityOfSections();
		setHScrollBarMode(QListView::Auto);
	}
}

void PlayListView::headerClicked(int section)
{
	if (!aspectSource_) return;

	DENTERMETHOD;
	if (section != aspectSource_->sortedColumn())
		aspectSource_->setSortParams(section, true);
	else
		aspectSource_->setSortParams(aspectSource_->sortedColumn(), !aspectSource_->isSortAscending());

	// aspectSource_->setSortParams(...) will cause aspectSource_->execute();

	DPRINTF("Header section %d clicked. Preparing playlist item index...", section);
	DEXITMETHOD;
}

void PlayListView::setSorting(int column, bool ascending)
{
	// omit sorting. We'll do it in our DB backend...
	// NOTE: This is a no-op stub, because the ListView widget calls it internally
}

void PlayListView::setPlayOrder(PlayOrder newOrder)
{
	currentIndex = 0;
	playOrder_ = newOrder;
	// generatePlayOrderTable();
}

/*!s
	@brief	generatePlayOrderTable()で使用

	libstdc++がリンクできないのでalgorithmのstd::swap()の代用
*/
void PlayListView::swap(int &n1, int &n2)
{
	int n = n1;
	n1 = n2;
	n2 = n;
}

/*!
	@brief	再生順序のテーブルを作成する

	アイテムの数が変化する場合は必ず呼び出す
*/
void PlayListView::generatePlayOrderTable()
{
	currentIndex = 0;

	delete playOrderTable;
	playOrderTable = NULL;

	const int count = count_ = this->childCount();

	if (count < 1)
		return;

	if (playOrder_ == ORDER_RANDOM || playOrder_ == ORDER_REPEAT_RANDOM)
	{
		playOrderTable = new int[count];

		for (int i = 0; i != count; ++i)
			playOrderTable[i] = i;

		// Randomize
		std::srand(time(NULL));

		for (int i = 0; i != count; ++i)
			swap(playOrderTable[i], playOrderTable[std::rand() % count]);

		if (currentItem())
		{
			int selectedIndex = itemIndex(currentItem());

			// make selected index the first in our play order table...
			for (int i = 0; i != count; ++i)
				if (playOrderTable[i] == selectedIndex)
				{
					swap(playOrderTable[i], playOrderTable[0]);
					break;
				}
		}
	}
}

void PlayListView::checkPlayOrderTable()
{
	if (!playOrderTable || count_ != this->childCount())
		generatePlayOrderTable();
}

Media *PlayListView::currentMedia()
{
	if (childCount() < 1)
		return NULL;

	if (playOrder_ == ORDER_RANDOM || playOrder_ == ORDER_REPEAT_RANDOM)
	{
		checkPlayOrderTable();
		setActiveItemByIndex(playOrderTable[currentIndex]);
	}
	else if (!activeItem_)
	{
		if (currentItem())
			setActiveItem(static_cast<PlayListViewItem *>(currentItem()));
		else
			setActiveItem(firstItem());
	}

	return activeItem_->media();
}

PlayListViewItem *PlayListView::firstItem()
{
	return static_cast<PlayListViewItem *>(firstChild());
}

PlayListViewItem *PlayListView::lastItem()
{
	if (childCount() < 1)
		return NULL;

	PlayListViewItem *item = activeItem_;

	if (!item)
		item = firstItem();

	if (!item)
		return NULL;

	while (item->nextSibling())
		item = static_cast<PlayListViewItem *>(item->nextSibling());

	return item;
}

PlayListViewItem *PlayListView::previousItem()
{
	if (!activeItem_)
		return NULL;

	if (activeItem_ == firstChild())
		return NULL;

	if (activeItem_->itemAbove())
		return static_cast<PlayListViewItem *>(activeItem_->itemAbove());
	else
		return NULL;
}

PlayListViewItem *PlayListView::nextItem()
{
	if (!activeItem_)
		return NULL;

	if (activeItem_->nextSibling())
		return static_cast<PlayListViewItem *>(activeItem_->nextSibling());
	else
		return NULL;
}

PlayListViewItem *PlayListView::nextPlayItem(bool modifyCurrentIndex)
{
	int count = this->childCount();

	if (count < 1)
		return NULL;

	PlayListViewItem* item = NULL;

	switch (playOrder_)
	{
	case ORDER_NORMAL:
		if (!activeItem_)
			item = static_cast<PlayListViewItem *>(currentItem());
		else if (!nextItem())
			item = NULL;
		else
			item = nextItem();

		break;

	case ORDER_REPEAT_ONE:
		item = activeItem_;

		break;

	case ORDER_REPEAT_ALL:
		if (!nextItem())
			item = firstItem();
		else
			item = nextItem();

		break;

	case ORDER_RANDOM:
		checkPlayOrderTable();

		if ((currentIndex + 1) < count)
			item = findItemByIndex(playOrderTable[modifyCurrentIndex ? ++currentIndex : currentIndex + 1]);
		else
			item = NULL;

		break;

	case ORDER_REPEAT_RANDOM:
		checkPlayOrderTable();

		if ((currentIndex + 1) < count)
			item = findItemByIndex(playOrderTable[modifyCurrentIndex ? ++currentIndex : currentIndex + 1]);
		else
			item = findItemByIndex(playOrderTable[modifyCurrentIndex ? currentIndex = 0 : 0]);

		break;
	}

	return item;
}

PlayListViewItem *PlayListView::previousPlayItem(bool modifyCurrentIndex)
{
	int count = this->childCount();

	if (count < 1)
		return NULL;

	PlayListViewItem* item = NULL;

	switch (playOrder_)
	{
	case ORDER_NORMAL:
		if (activeItem_ == firstItem())
			item = NULL;
		else
			item = previousItem();

		break;

	case ORDER_REPEAT_ONE:
		break;

	case ORDER_REPEAT_ALL:
		if (activeItem_ == firstItem())
			item = lastItem();
		else
			item = previousItem();

		break;

	case ORDER_RANDOM:
		checkPlayOrderTable();

		if (currentIndex <= 0)
			item = NULL;
		else
			item = findItemByIndex(playOrderTable[modifyCurrentIndex ? --currentIndex : currentIndex - 1]);

		break;

	case ORDER_REPEAT_RANDOM:
		checkPlayOrderTable();

		if (currentIndex <= 0)
			item = NULL;
		else
			item = findItemByIndex(playOrderTable[modifyCurrentIndex ? --currentIndex : currentIndex - 1]);

		break;
	}

	return item;
}

Media *PlayListView::nextMedia()
{
	PlayListViewItem *item = nextPlayItem(true);

	if (!item)
		return NULL;

	setActiveItem(item, true, playOrder_ == ORDER_RANDOM || playOrder_ == ORDER_REPEAT_RANDOM);
	return currentMedia();
}

Media *PlayListView::previousMedia()
{
	PlayListViewItem *item = previousPlayItem(true);

	if (!item)
		return NULL;

	setActiveItem(item, true, playOrder_ == ORDER_RANDOM || playOrder_ == ORDER_REPEAT_RANDOM);
	return currentMedia();
}

/*
PlayListViewItem *PlayListView::insertItem(unsigned long mediaID, PlayListViewItem *after)
{
	if (mediaID)
	{
		if (!after)
		{
			QListViewItemIterator it(this);
			QListViewItemIterator last;
			while (it.current()) {
				last = it;
				++ it;
			}

			after = (PlayListViewItem *)last.current();
		}

		// add to database...
		sqlite3_stmt *vm = stmtInsertMediaIDIntoPlaylist;
		sqlite3_bind_null(vm, 1);
		sqlite3_bind_null(vm, 2);
		sqlite3_bind_int(vm, 3, mediaID);
		sqlite3_step(vm);
		sqlite3_reset(vm);
		// ... and get new playlist ID
		int newID = sqlite3_last_insert_rowid(mediaDatabase_->db());

		if (!aspectSource_->isGraphFiltered() && after)
			dbPlaylistItemOrderDirty_ = true;

		return new PlayListViewItem(mediaID, newID, this, after);


	}
	else
		return NULL;
}

PlayListViewItem *PlayListView::insertItem(const QString &mediaFilename, PlayListViewItem *after)
{
	unsigned long mediaID = mediaDatabase_->getMediaIDForLocation(mediaFilename);

	if (mediaID)
		return insertItem(mediaID, after);
	else
		return NULL;
}
*/

void PlayListView::keyPressEvent(QKeyEvent *e)
{
	if (e->key() == Qt::Key_Delete && currentItem())
	{
		takeItem(currentItem());
		e->accept();
		return;
	}

	QListView::keyPressEvent(e);
}

void PlayListView::updateResizabilityOfSections()
{
	for (int i = 0; i < columns(); ++i)
		header()->setResizeEnabled(true, header()->mapToSection(i));

	header()->setResizeEnabled(false, 0);

	if (autoScaleColumnsEnabled_)
		header()->setResizeEnabled(false, header()->mapToSection(columns() - 1));
}

void PlayListView::setAutoSectionSizes()
{
	if (autoResizingSections_)
		return;

	autoResizingSections_ = true;

	if (headerSectionSizes_.count() == 0)
		updateDynamicSectionSizes();

	float totalSectionsWidth = header()->width();
	totalSectionsWidth -= header()->sectionSize(0);
	totalSectionsWidth -= header()->sectionSize(1);
	totalSectionsWidth -= header()->sectionSize(5);

	setColumnWidth(2, (int)(totalSectionsWidth * headerSectionSizes_[0] / 100));
	setColumnWidth(3, (int)(totalSectionsWidth * headerSectionSizes_[1] / 100));
	setColumnWidth(4, (int)(totalSectionsWidth * headerSectionSizes_[2] / 100));

	updateResizabilityOfSections();

	autoResizingSections_ = false;
}

void PlayListView::resizeEvent(QResizeEvent *e)
{
	QListView::resizeEvent(e);

	if (autoScaleColumnsEnabled_)
		setAutoSectionSizes();
}

void PlayListView::contentsMousePressEvent(QMouseEvent *e)
{
	if (inputMode() == PlayListView::Move)
	{
		buttonDown_ = true;
		moveTriggered_ = false;
		buttonDownPos_ = e->pos();
	}
	else
	{
		QListView::contentsMousePressEvent(e);

		if (e->button() == Qt::RightButton && popupMenu_)
			popupMenu_->exec(e->globalPos());
	}
}

void PlayListView::contentsMouseMoveEvent(QMouseEvent *e)
{
	if (inputMode() == PlayListView::Move)
	{
		if (buttonDown_ && !aspectSource_->isGraphFiltered())
		{
			if (moveTriggered_)
			{
			    QPoint vp = contentsToViewport(e->pos());
			    QListViewItem *i = itemAt(vp);
			    bool moveFirstChild = false;

			    if (i == firstChild())
			    {
			    	moveFirstChild = true;
			    }
			    else if (i)
			    {
			    	QListViewItem *iabove = i->itemAbove();
			    	if (iabove && !selectedItemsForMove_.contains(static_cast<PlayListViewItem *>(iabove)))
			    		i = iabove;
			    }
			    else
			    {
			    	i = firstChild();
			    	while (i->nextSibling())
			    		i = i->nextSibling();
			    }

			    if (!selectedItemsForMove_.contains(static_cast<PlayListViewItem *>(i)))
			    {
				    PlayListViewItem *pli;

				    for (int j = 0; j < selectedItemsForMove_.count(); ++i)
				    {
				    	pli = selectedItemsForMove_.at(j);
				    	pli->moveItem(i);
				    	i = pli;
				    }

				    if (moveFirstChild)
				    	firstChild()->moveItem(i);

				    dbPlaylistItemOrderDirty_ = true;
			    }
			}
			else
			{
				if (e->pos().y() > buttonDownPos_.y() + 10 ||
					e->pos().y() < buttonDownPos_.y() - 10)
				{
					moveTriggered_ = true;
					selectedItemsForMove_ = selectedItems();
				}
			}

		    return;
		}
	}

	QListView::contentsMouseMoveEvent(e);
}

void PlayListView::contentsMouseReleaseEvent(QMouseEvent *e)
{
	buttonDown_ = false;
	moveTriggered_ = false;

	QListView::contentsMouseReleaseEvent(e);
}

PlayListViewItem *PlayListView::findItemByID(unsigned long playlistID)
{
	QListViewItemIterator it(this);
	for (; it.current() && static_cast<PlayListViewItem *>(it.current())->id() != playlistID; ++it)
		;

	return static_cast<PlayListViewItem *>(it.current());
}

bool PlayListView::setActiveItemByID(unsigned long playlistID)
{
	DHPRINTF("setActiveItemByID %d", playlistID);

	PlayListViewItem *foundItem = findItemByID(playlistID);

	if (foundItem)
	{
		setActiveItem(foundItem, true);
		return true;
	}

	return false;
}
PlayListViewItem *PlayListView::findItemByMediaID(unsigned long mediaID)
{
	QListViewItemIterator it(this);
	for (; it.current() && static_cast<PlayListViewItem *>(it.current())->mediaID() != mediaID; ++it)
		;

	return static_cast<PlayListViewItem *>(it.current());
}

bool PlayListView::setActiveItemByMediaID(unsigned long mediaID)
{
	DHPRINTF("setActiveItemByMediaID %d", mediaID);

	PlayListViewItem *foundItem = findItemByMediaID(mediaID);

	if (foundItem)
	{
		setActiveItem(foundItem, true);
		return true;
	}

	return false;
}

PlayListViewItem *PlayListView::findItemByIndex(int index)
{
	if (this->childCount() <= index)
		return NULL;

	QListViewItemIterator it(this);
	for (int i = 0; it.current() && i != index; ++it, ++i);

	return static_cast<PlayListViewItem *>(it.current());
}

void PlayListView::setActiveItemByIndex(int index)
{
	PlayListViewItem *foundItem = findItemByIndex(index);

	if (foundItem)
		setActiveItem(foundItem, true, true); // also select new item (ie. in random mode...)
}

void PlayListView::setActiveItem(PlayListViewItem *item, bool ensureVisible, bool selectItem)
{
	PlayListViewItem *oldActiveItem = activeItem_;
	activeItem_ = item;

	if (activeItem_)
	{
		activeID_ = item->id();
		activeMediaID_ = item->mediaID();
	}
	else
	{
		activeID_ = 0;
		activeMediaID_ = 0;
	}

	if (oldActiveItem)
		oldActiveItem->repaint();

	if (activeItem_)
	{
		activeItem_->repaint();

		if (selectItem)
		{
			setCurrentItem(activeItem_);
			setSelected(activeItem_, true);
		}

		if (ensureVisible)
			ensureItemVisible(activeItem_);
	}
}

void PlayListView::setActiveItemPaused(bool y)
{
	activeItemPaused_ = y;
	if (activeItem_)
		activeItem_->repaint();
}

void PlayListView::setCurrentItemActive()
{
	setActiveItem(static_cast<PlayListViewItem *>(currentItem()));
	setActiveItemPaused(false);
}

int PlayListView::itemIndex(QListViewItem* item)
{
	if (this->childCount() == 0)
		return -1;

	int i = 0;
	for (QListViewItemIterator it(this); it.current() && it.current() != item; ++it, ++i)
		;

	return i;
}

void PlayListView::saveColumnOrder(const QString &configName, const QString &groupName)
{
	Config cfg(configName, Config::File);
	cfg.setGroup(groupName);

	for (int i = 0; i < columns(); ++i)
	{
		int section = header()->mapToSection(i);
		cfg.writeEntry(
			"section_" + QString::number(i),
			section != aspectSource_->sortedColumn() ?
					QString::number(section) :
					QString("%1,%2").arg(section).arg(aspectSource_->isSortAscending() ? 1 : 0)
		);
	}
}

void PlayListView::saveColumnWidth(const QString &configName, const QString &groupName)
{
	Config cfg(configName, Config::File);
	cfg.setGroup(groupName);

	for (int i = 0; i < columns(); ++i)
		cfg.writeEntry("width_" + QString::number(i), columnWidth(header()->mapToSection(i)));
}

void PlayListView::loadColumnOrder(const QString &configName, const QString &groupName)
{
	Config cfg(configName, Config::File);
	cfg.setGroup(groupName);

	for (int i = 0; i < columns(); ++i)
	{
		QString key = "section_" + QString::number(i);
		if (cfg.hasKey(key))
		{
			QStringList args = cfg.readListEntry(key, ',');
			int section = args[0].toInt();
			header()->moveSection(section, i);
			if (args.count() == 2)
				aspectSource_->setSortParams(section, args[1].toInt() == 1);
		}
	}

	updateResizabilityOfSections();
}

void PlayListView::loadColumnWidth(const QString &configName, const QString &groupName)
{
	Config cfg(configName, Config::File);
	cfg.setGroup(groupName);

	for (int i = 0; i < columns(); ++i)
	{
		QString key = "width_" + QString::number(i);
		if (cfg.hasKey(key))
			setColumnWidth(header()->mapToSection(i), QMAX(10, cfg.readNumEntry(key)));
	}

	setColumnWidth(0, 20);

	if (autoScaleColumnsEnabled_)
		setAutoSectionSizes();
}

void PlayListView::setBackgroundColors(const QColor &evenColor, const QColor &oddColor)
{
	oddRowColorGroup_.setColor(QColorGroup::Base, oddColor);
	evenRowColorGroup_.setColor(QColorGroup::Base, evenColor);
}

void PlayListView::setForegroundColor(const QColor &textColor)
{
	oddRowColorGroup_.setColor(QColorGroup::Text, textColor);
	oddRowColorGroup_.setColor(QColorGroup::Foreground, textColor);
	evenRowColorGroup_.setColor(QColorGroup::Text, textColor);
	evenRowColorGroup_.setColor(QColorGroup::Foreground, textColor);
}

void PlayListView::setPlayIndicatorImage(const QPixmap &image)
{
	playIndicatorImage_ = &image;
}

void PlayListView::setPauseIndicatorImage(const QPixmap &image)
{
	pauseIndicatorImage_ = &image;
}

void PlayListView::setErrorIndicatorImage(const QPixmap &image)
{
	errorIndicatorImage_ = &image;
}

void PlayListView::setSkinMode(const SkinModeInformation &modeInfo)
{
	if (modeInfo.playListFont().font.isUsed())
		setFont(modeInfo.playListFont().font);
	else
		unsetFont();

	// set general palette:
	QPalette palette = QApplication::palette();

	if (modeInfo.playListFont().fontColor.isUsed())
	{
		palette.setColor(QColorGroup::Text, modeInfo.playListFont().fontColor);
		palette.setColor(QColorGroup::Foreground, modeInfo.playListFont().fontColor);
	}

	if (modeInfo.playListBackgroundColorSelected().isUsed())
	{
		palette.setColor(QColorGroup::Highlight, modeInfo.playListBackgroundColorSelected());
		palette.setColor(QColorGroup::HighlightedText, modeInfo.playListFontColorSelected());
	}

	// set even and odd colors for items:
	if (modeInfo.playListBackgroundColorEven().isUsed())
	{
		if (modeInfo.playListBackgroundColorOdd().isUsed())
			setBackgroundColors(modeInfo.playListBackgroundColorEven(), modeInfo.playListBackgroundColorOdd());
		else
			setBackgroundColors(modeInfo.playListBackgroundColorEven(), modeInfo.playListBackgroundColorEven());

		palette.setColor(QColorGroup::Base, modeInfo.playListBackgroundColorEven());
	}
	else
		setBackgroundColors(palette.active().base(), palette.active().base());

	setPalette(palette);

	setPlayIndicatorImage(modeInfo.playListPlayIndicatorImage());
	setPauseIndicatorImage(modeInfo.playListPauseIndicatorImage());
	setErrorIndicatorImage(modeInfo.playListErrorIndicatorImage());

	if (modeInfo.playListHeaderFont().font.isUsed())
		header()->setFont(modeInfo.playListHeaderFont().font);
	else
		header()->unsetFont();

	if (modeInfo.playListHeaderColor().isUsed())
	{
		QPalette palette(modeInfo.playListHeaderColor());

		if (modeInfo.playListHeaderFont().fontColor.isUsed())
		{
			palette.setColor(QColorGroup::Text, modeInfo.playListHeaderFont().fontColor);
			palette.setColor(QColorGroup::ButtonText, modeInfo.playListHeaderFont().fontColor);
			palette.setColor(QColorGroup::Foreground, modeInfo.playListHeaderFont().fontColor);
		}

		header()->setPalette(palette);
	}
	else
		header()->unsetPalette();

	if (modeInfo.playListScrollBarColor().isUsed())
	{
		horizontalScrollBar()->setPalette(QPalette(modeInfo.playListScrollBarColor()));
		verticalScrollBar()->setPalette(QPalette(modeInfo.playListScrollBarColor()));
	}
	else
	{
		horizontalScrollBar()->unsetPalette();
		verticalScrollBar()->unsetPalette();
	}
}

void PlayListView::cleanUpMediaCache()
{
	mediaDatabase_->compactDB();
}

void PlayListView::saveMediaCache()
{
	if (mediaDatabase_->hasChanged())
	{
		DPRINTF("Saving media cache...");
		mediaDatabase_->save();
	}
}

/*
void PlayListView::setPlayListSource(const QString &name)
{
	if (name != state.playListSource)
	{
		// We are about to switch the playlist source table.
		// So, make sure the current order is saved to the table...
		ensurePlayListValidInDB();
		state.playListSource = name;
		mediaDatabaseFinalizingStatements(mediaDatabase_->db());
		mediaDatabasePreparingStatements(mediaDatabase_->db());
		reloadView();
	}
}

void PlayListView::setViewSource(const QString &name)
{
	DENTERMETHODF("PlayListView::setViewSource(\"%s\")", (const char*)name.toUtf8());
	// We are about to switch the view source table.
	// So, make sure the current order is saved to the table...
	ensurePlayListValidInDB();
	viewSource_ = name;
	updateView();
	DEXITMETHODF("PlayListView::setViewSource(\"%s\")", (const char*)name.toUtf8());
}
*/

void PlayListView::mediaDatabasePreparingStatements(sqlite3 *db)
{
	QString query = "INSERT INTO " + aspectSource_->playListAspect()->playList()->sourceTable() + " VALUES(?1, ?2, ?3);";
	sqlite3_prepare_v2(db, (const char *)query.toUtf8(), -1, &stmtInsertMediaIDIntoPlaylist, 0);
}

void PlayListView::mediaDatabaseFinalizingStatements(sqlite3 *db)
{
	sqlite3_finalize(stmtInsertMediaIDIntoPlaylist);
	stmtInsertMediaIDIntoPlaylist = NULL;
}

void PlayListView::itemSelectedStateChanged(PlayListViewItem *item, bool selected)
{
	DPRINTF("==============================");

	if (selected)
	{
		DPRINTF("item selected %d", (uint)item);
		selectedItems_.append(item);
	}
	else
	{
		DPRINTF("item deselected %d", (uint)item);
		selectedItems_.remove(item);
	}

	DPRINTF("selectedItems_.count() = %d", selectedItems_.count());
	DPRINTF("-----------------------------");
}
